'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _json2md = require('json2md');

var _json2md2 = _interopRequireDefault(_json2md);

var _pgStructure = require('pg-structure');

var _pgStructure2 = _interopRequireDefault(_pgStructure);

var _pg = require('pg');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const d = new _debug2.default('make-postgres-markdown');
const functionNameFromActionStatement = /execute procedure (.*)\s*\(/i;

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (options) {
    console.time('make-postgres-markdown');
    d('Parsing schema');

    const ignore = options.ignore ? new RegExp(options.ignore) : false;

    const client = new _pg.Client(options);
    client.connect();

    d('Triggers');
    const triggers = yield client.query(`
SELECT *
from information_schema.triggers
WHERE event_object_schema = '${options.schema}'
  `);

    // get all functions referenced by a trigger
    const knownFunctions = [];
    triggers.rows.forEach(function (trigger) {
      const match = functionNameFromActionStatement.exec(trigger.action_statement);
      if (match.length > 1) {
        knownFunctions.push(`'${match[1]}'`);
      }
    });

    d('Building JSON from manual queries...');
    d('Functions');
    const functions = yield client.query(`
SELECT *
FROM pg_catalog.pg_proc
WHERE proowner != 10
OR proname IN (${knownFunctions.join(',')})
  `);

    d('Extensions');
    const extensions = yield client.query(`
SELECT *
FROM pg_available_extensions
WHERE installed_version IS NOT null;
  `);

    client.end();

    const db = yield (0, _pgStructure2.default)(options, [options.schema]);
    d('Building JSON representation from pg-structure...');
    const schema = db.schemas.get('public');
    const tables = schema.tables;

    const markdown = [{ h1: 'Tables' }, ...renderTables(tables, 'table', ignore), { h1: 'Views' }, ...renderTables(tables, 'view', ignore)];

    if (functions.rows.length) {
      markdown.push({
        h1: 'Functions'
      });

      functions.rows.forEach(function (func) {
        markdown.push({
          h2: func.proname
        });
        markdown.push({
          p: func.prosrc
        });
      });
    }

    if (extensions.rows) {
      markdown.push({
        h1: 'Extensions'
      });

      markdown.push({
        table: {
          headers: ['name', 'version', 'description'],
          rows: extensions.rows.map(function (extension) {
            return {
              name: extension.name,
              version: extension.installed_version,
              description: extension.comment
            };
          })
        }
      });
    }

    d('Converting JSON to markdown');
    const output = (0, _json2md2.default)(markdown);

    d('Writing output');
    _fs2.default.writeFileSync(options.output, `---
title: Database Documentation

search: true
---

${output}
  `);

    d('Finished');
    console.timeEnd('make-postgres-markdown');

    function renderConstraints(column) {
      const constraints = [];

      if (!column.allowNull) {
        constraints.push('NOT NULL');
      }

      for (let [constraintName, constraint] of column.foreignKeyConstraints) {
        for (let [name, column] of constraint.columns) {
          constraints.push(`[${name}](#${constraint.referencedTable.name})`);
        }
      }

      return constraints.length && constraints.join(', ');
    }

    function renderTables(tables, kind, ignore) {
      const markdown = [];

      d(`rendering tables (${kind})`);
      for (let [name, table] of tables) {
        if (table.kind !== kind || ignore && ignore.exec(name)) {
          continue;
        }

        markdown.push({ h2: name });
        const markdownTable = {
          headers: ['column', 'comment', 'type', 'default', 'constraints', 'values'],
          rows: []
        };

        for (let [name, column] of table.columns) {
          markdownTable.rows.push([name || '', column.comment || '', column.type || '', column.default || '', renderConstraints(column) || '', column.enumValues ? column.enumValues.join(', ') : '']);
        }

        markdown.push({ table: markdownTable });

        const tableTriggers = triggers.rows.filter(trigger => trigger.event_object_table === name).map(trigger => ({
          name: trigger.trigger_name,
          timing: trigger.action_timing,
          orientation: trigger.action_orientation,
          manipulation: trigger.event_manipulation,
          statement: actionStatementToFunctionLink(trigger.action_statement)
        }));

        if (tableTriggers.length) {
          markdown.push({ h3: 'Triggers' });
          markdown.push({
            table: {
              headers: ['name', 'timing', 'orientation', 'manipulation', 'statement'],
              rows: tableTriggers
            }
          });
        }
      }

      return markdown;
    }
  });

  function makeMarkdown(_x) {
    return _ref.apply(this, arguments);
  }

  return makeMarkdown;
})();

function actionStatementToFunctionLink(actionStatement) {
  const functionName = functionNameFromActionStatement.exec(actionStatement);
  if (functionName.length > 1) {
    return actionStatement.replace(functionName[1], `<a href="#${functionName[1]}">${functionName[1]}</a>`);
  }

  return actionStatement;
}